"""Common utility functions."""
import datetime as dt
import json
import pathlib
import sys
import platform
import subprocess
import webbrowser
import itertools
import re
import dateutil.parser
import pick
try:
    import gnureadline as readline
except ImportError:
    import readline


# Map semester name to number
SEMESTER_NUM = {"Winter": 1, "Spring": 2, "Summer": 3, "Fall": 4}

# Map month number to semester number
MONTH_SEMESTER_NUM = {
    1: 1, 2: 1, 3: 1, 4: 1,     # Jan-Apr Winter
    5: 2, 6: 2,                 # May-Jun Spring
    7: 3, 8: 3,                 # Jul-Aug Summer
    9: 4, 10: 4, 11: 4, 12: 4,  # Sep-Dec Fall
}


def dict_str(obj):
    """Format a dictionary as an indented string."""
    return json.dumps(obj, indent=4)


def course_str(course):
    """Format course as a string."""
    return (
        f"[{course['pk']}] {course['name']} "
        f"{course['semester']} {course['year']}"
    )


def course_key(course):
    """Return a tuple for sorting courses by year, semester, and name."""
    # Coerce year
    if course["year"] is None:
        year = 0
    else:
        year = course["year"]

    # Convert semester to a number
    if course["semester"] is None:
        semester_num = 0
    else:
        semester_num = SEMESTER_NUM[course["semester"]]

    # Coerce name
    name = course["name"]
    if name is None:
        name = ""

    # Return a tuple in order of sort precedence
    return (year, semester_num, name)


def is_current_course(course):
    """Return True if course is from current or future semester."""
    # Coerce year
    if course["year"] is None:
        year = 0
    else:
        year = course["year"]

    # Convert semester to a number
    if course["semester"] is None:
        semester_num = 0
    else:
        semester_num = SEMESTER_NUM[course["semester"]]

    # Compare course year and semester to today
    today = dt.date.today()
    return (
        year >= today.year and
        semester_num >= MONTH_SEMESTER_NUM[today.month]
    )


def course_match(search, courses):
    """Return courses matching search term."""
    year, semester, name = parse_course_string(search)
    courses = filter(
        lambda x:
            x["year"] == year and
            x["semester"] == semester and
            name in x["name"],
        courses
    )
    return list(courses)


def parse_course_string(user_input):
    """Return year, semester, and course name from a user input string.

    EXAMPLE:
    >>> parse_course_string("eecs485sp21")
    (2021, 'Spring', 'EECS 485')
    """
    pattern = r"""(?ix)         # Regex options: case insensitive, verbose
    ^                           # Match starts at beginning
    [\s]*                       # Optional leading whitespace
    (?P<dept>[a-z]*)            # Optional department
    [\s_-]*                     # Optional whitespace or delimiter
    (?P<num>\d{3})              # 3 digit course number
    [\s_-]*                     # Optional whitespace or delimiter
    (?P<sem>                    # Semester name or abbreviation
        w|wn|winter|
        sp|s|spring|
        su|summer|
        sp/su|spsu|ss|spring/summer|
        f|fa|fall)
    [\s_-]*                     # Optional whitespace or delimiter
    (?P<year>\d{2,4})           # 2-4 digit year
    \s*                         # Optional trailing whitespace
    $                           # Match ends at the end
    """
    match = re.search(pattern, user_input)
    if not match:
        sys.exit(f"Error: unsupported input format: '{user_input}'")

    # Convert year to a number, handling 2-digit year as "20xx"
    year = int(match.group("year"))
    assert year >= 0
    if year < 100:
        year = 2000 + year

    # Convert semester abbreviation to semester name.  Make sure that the keys
    # match the abbreviations in the regular expression above.
    semester_names = {
        "w": "Winter",
        "wn": "Winter",
        "winter": "Winter",
        "sp": "Spring",
        "s": "Spring",
        "spring": "Spring",
        "su": "Summer",
        "summer": "Summer",
        "sp/su": "Spring/Summer",
        "spsu": "Spring/Summer",
        "ss": "Spring/Summer",
        "spring/summer": "Spring/Summer",
        "f": "Fall",
        "fa": "Fall",
        "fall": "Fall",
    }
    semester_abbrev = match.group("sem").lower()
    semester = semester_names[semester_abbrev]

    # Course name, with department and catalog number.  If no department is
    # specified, assume "EECS".
    dept = match.group("dept")
    dept = dept.upper()
    if not dept:
        print("WARNING: no department found, assuming 'EECS'", file=sys.stderr)
        dept = "EECS"
    num = match.group("num")
    name = f"{dept} {num}"

    # Return a tuple
    return year, semester, name


def get_current_course_list(client):
    """Return a sorted list of current and future courses."""
    user = client.get("/api/users/current/")
    courses = client.get(f"/api/users/{user['pk']}/courses_is_admin_for/")
    courses += client.get(f"/api/users/{user['pk']}/courses_is_staff_for/")
    courses = sorted(courses, key=course_key, reverse=True)
    courses = [k for k, v in itertools.groupby(courses)]  # Unique
    return courses


def get_course_smart(course_arg, client):
    """Interact with the user to select a course.

    1. If course_arg is a number, look up course by primary key
    2. If course_arg is None, prompt with a list of current or futures courses
    3. If course_arg is a string, try to extract semester, year and name, then
       match against list of courses for which user is an admin.

    This function provides sanity checks and may exit with an error message.
    """
    # User provides course PK
    if course_arg and course_arg.isnumeric():
        return client.get(f"/api/courses/{course_arg}/")

    # Get a list of courses sorted by year, semester and name
    courses = get_current_course_list(client)

    # No course input from the user.  Show all courses and prompt the user.
    if not course_arg:
        if not courses:
            sys.exit("Error: No current courses, try 'agio courses -l'")
        else:
            selected_courses = pick.pick(
                options=courses,
                title=("Select a course:"),
                options_map_func=course_str,
                multiselect=False,
            )
            assert selected_courses
            return selected_courses[0]

    # Try to match a course
    matches = course_match(course_arg, courses)
    if not matches:
        courses_str = "\n".join(course_str(i) for i in courses)
        sys.exit(
            f"Error: no course matches '{course_arg}'\n"
            f"{courses_str}"
        )
    elif len(matches) > 1:
        matches_str = "\n".join(course_str(i) for i in matches)
        sys.exit(
            f"Error: multiple courses match '{course_arg}'\n"
            f"{matches_str}"
        )
    return matches[0]


def parse_project_string(user_input):
    """Return assignment type, number, and subtitle from a user input string.

    EXAMPLE:
    >>> parse_project_string("p4mapreduce")
    ('Project', 4, 'mapreduce')
    """
    pattern = r"""(?ix)         # Regex options: case insensitive, verbose
    ^                           # Match starts at beginning
    \s*                         # Optional leading whitespace
    (?P<asstype>[a-z]*)         # Assignment type
    [\s_-]*                     # Optional whitespace or delimiter
    (?P<num>[\d]*)              # Assignment number
    [\s_-]*                     # Optional whitespace or delimiter
    (?P<subtitle>.*)            # Assignment subtitle
    \s*                         # Optional trailing whitespace
    $                           # Match ends at the end
    """
    assert isinstance(user_input, str)
    match = re.search(pattern, user_input)
    if not match:
        sys.exit(f"Error: unsupported input format: '{user_input}'")

    # HACK: If the input is just a string, e.g., "images", then it will be
    # mis-classified as asstype.
    if (match.group("asstype") and not match.group("num")
            and not match.group("subtitle")):
        subtitle = match.group("asstype")
        return None, None, subtitle

    # Convert assignment type abbreviations to canonical "Project", "Lab", etc.
    assignment_types = {
        "p": "Project",
        "proj": "Project",
        "project": "Project",
        "l": "Lab",
        "lab": "Lab",
        "h": "Homework",
        "hw": "Homework",
        "hwk": "Homework",
        "homework": "Homework",
    }
    asstype_abbrev = match.group("asstype").lower()
    if asstype_abbrev not in assignment_types:
        asstypes = ", ".join(assignment_types.keys())
        sys.exit(
            f"Error: unsupported assignment type: '{asstype_abbrev}'.  "
            f"Recognized shortcuts: {asstypes}"
        )
    asstype = assignment_types[asstype_abbrev]

    # Assignment number and subtitle
    num = int(match.group("num"))
    subtitle = match.group("subtitle")

    return asstype, num, subtitle


def project_str(project):
    """Format project as a string."""
    return f"[{project['pk']}] {project['name']}"


def project_match(search, projects):
    """Return projects matching search term."""
    assert projects
    asstype, num, subtitle = parse_project_string(search)

    # Remove projects with an assignment type mismatch (Lab vs. Project, etc.)
    if asstype:
        projects = filter(
            lambda x:
                parse_project_string(x["name"])[0].lower() == asstype.lower(),
            projects
        )

    # Remove projects with a number mismatch
    if num:
        projects = filter(
            lambda x: parse_project_string(x["name"])[1] == num,
            projects
        )

    # Remove projects with a name mismatch, tolerating substring match
    if subtitle:
        projects = filter(
            lambda x:
                subtitle.lower() in parse_project_string(x["name"])[2].lower(),
            projects
        )

    return list(projects)


def get_course_project_list(course, client):
    """Return a sorted list of projects for course."""
    projects = client.get(f"/api/courses/{course['pk']}/projects/")
    projects = sorted(projects, key=lambda x: x["name"])
    return projects


def get_project_smart(project_arg, course_arg, client):
    """Interact with the user to select a project.

    1. If project_arg is a number, look up project by primary key
    2. Use previously defined procedure for smart course selection
    3. If project_arg is None, prompt with list of projects for selected course
    4. If project_arg is a string, try to match project name

    This function provides sanity checks and may exit with an error message.
    """
    # User provides project PK
    if project_arg and project_arg.isnumeric():
        return client.get(f"/api/projects/{project_arg}/")

    # Get a course and a sorted list of projects
    course = get_course_smart(course_arg, client)
    projects = get_course_project_list(course, client)
    if not projects:
        sys.exit("Error: No projects for course, try 'agio courses -l'")

    # No project input from the user.  Show all projects for current course and
    # and prompt the user.
    if not project_arg:
        selected_projects = pick.pick(
            options=projects,
            title="Select a project:",
            options_map_func=project_str,
            multiselect=False,
        )
        assert selected_projects
        return selected_projects[0]

    # User provides strings, try to match a project
    matches = project_match(project_arg, projects)
    if not matches:
        projects_str = "\n".join(project_str(i) for i in projects)
        sys.exit(
            f"Error: no project matches '{project_arg}'\n"
            f"{projects_str}"
        )
    elif len(matches) > 1:
        matches_str = "\n".join(project_str(i) for i in matches)
        sys.exit(
            f"Error: multiple projects match '{project_arg}'\n"
            f"{matches_str}"
        )
    return matches[0]


def group_str(group):
    """Format group as string."""
    uniqnames = group_uniqnames(group)
    uniqnames_str = ", ".join(uniqnames)
    return f"[{group['pk']}] {uniqnames_str}"


def group_uniqnames(group):
    """Return group member uniqnames."""
    members = group["members"]
    return [x["username"].replace("@umich.edu", "") for x in members]


def is_group_member(uniqname, group):
    """Return True if uniqname is in group."""
    return uniqname in group_uniqnames(group)


def group_match(uniqname, groups):
    """Return groups where uniqname is a member."""
    matches = filter(lambda x: is_group_member(uniqname, x), groups)
    return list(matches)


def get_group_list(project, client):
    """Return a sorted list of groups for project."""
    groups = client.get(f"/api/projects/{project['pk']}/groups/")
    groups = sorted(groups, key=lambda x: x["pk"])
    return groups


def get_group_smart(group_arg, project_arg, course_arg, client):
    """Interact with the user to select a group.

    1. If group_arg is a number, look up group by primary key
    2. Use previously defined procedure for smart project selection, which in
       turn may use the smart course selection procedure.
    3. If group_arg is None, prompt with list of groups for selected project
    4. If group_arg is a string, try to match it to a group member uniqname

    This function provides sanity checks and may exit with an error message.
    """
    # User provides group PK
    if group_arg and group_arg.isnumeric():
        return client.get(f"/api/groups/{group_arg}/")

    # Get a project and list of groups
    project = get_project_smart(project_arg, course_arg, client)
    groups = get_group_list(project, client)
    if not groups:
        sys.exit("Error: No groups for project, try 'agio projects -l'")

    # No group input from the user.  Help them select a uniqname with an
    # auto complete prompt.
    if not group_arg:
        # Get a list of uniqnames
        uniqnames = set()
        for group in groups:
            uniqnames.update(group_uniqnames(group))

        # Register the completer function
        def uniqname_completer(text, state):
            """Complete a uniqname, callback run by readline.

            https://stackoverflow.com/questions/187621/how-to-make-a-python-command-line-program-autocomplete-arbitrary-things-not-int/187660#187660
            """
            options = [i for i in uniqnames if i.startswith(text)]
            if state < len(options):
                return options[state]
            return None
        readline.set_completer(uniqname_completer)

        # Use the tab key for completion
        readline.parse_and_bind('tab: complete')

        # Prompt the user to select a uniqname
        while True:
            uniqname = input("Uniqname (TAB to autocomplete): ")
            assert uniqname
            uniqname = uniqname.strip()
            group_arg = uniqname
            break

    # Try to match uniqname to a group member
    matches = group_match(group_arg, groups)
    if not matches:
        sys.exit(f"Error: uniqname not in any group: {group_arg}")
    elif len(matches) > 1:
        matches_str = "\n".join(group_str(i) for i in matches)
        sys.exit(
            f"Error: uniqname in more than one group: {group_arg}"
            f"{matches_str}"
        )
    return matches[0]


def is_wsl():
    """Check if user is running WSL."""
    return 'microsoft' in platform.uname().release


def open_web(url):
    """Open url in default browser."""
    if is_wsl():
        # Need to escape & in Windows
        # https://stackoverflow.com/questions/1327431/how-do-i-escape-ampersands-in-batch-files
        url = url.replace('&', '^&')
        subprocess.run(
            ['cmd.exe', '/c', 'start', url],
            stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL,
            check=True
        )
    else:
        webbrowser.open(url)


def submission_key(submission):
    """Return a tuple for sorting submissions by timestamp."""
    return dateutil.parser.parse(submission["timestamp"])


def submission_str(submission):
    """Format submission as a string."""
    timestamp = dateutil.parser.parse(submission["timestamp"])
    timestamp_human = timestamp.strftime("%Y-%m-%d %H:%M:%S")
    return (
        f"[{submission['pk']}] {timestamp_human}  "
    )


def get_submission_list(group, client):
    """Return a sorted list of submissions for a group."""
    submissions = client.get(f"/api/groups/{group['pk']}/submissions/")
    submissions = sorted(submissions, key=submission_key, reverse=True)
    return submissions


def get_submission_smart(
        submission_arg, group_arg, project_arg, course_arg, client):
    """Interact with the user to select a submission.

    1. If submission_arg is a number, look up submission by primary key
    2. Use previously defined procedure for smart group selection, which in
       turn may use the smart project and course selection procedures.
    3. If submission_arg is "best", show submission with highest score
    4. If submission_arg is "last", show most recent submission
    5. If submission_arg is None, prompt with list of submissions for the
       selected group

    This function provides sanity checks and may exit with an error message.
    """
    # User provides submission PK
    if submission_arg and submission_arg.isnumeric():
        return client.get(f"/api/submissions/{submission_arg}/")

    # Get a group
    group = get_group_smart(group_arg, project_arg, course_arg, client)

    # User provides "best"
    if submission_arg == "best":
        return client.get(f"/api/groups/{group['pk']}/ultimate_submission/")

    # Get a sorted list of submissions
    submissions = get_submission_list(group, client)
    if not submissions:
        sys.exit("Error: No submissions, try 'agio submissions -l'")

    # No submissions input from the user.  Show all submissions for this group
    # and prompt the user.
    if not submission_arg:
        selected_submissions = pick.pick(
            options=submissions,
            title="Select a submission:",
            options_map_func=submission_str,
            multiselect=False,
        )
        assert selected_submissions
        return selected_submissions[0]

    # User provides string "last"
    if submission_arg == "last":
        return submissions[0]

    # No other attempt to match
    submissions_str = "\n".join(submission_str(i) for i in submissions)
    sys.exit(
        f"Error: no submission matches '{submission_arg}'.  "
        "Provide a primary key or 'last'.\n"
        f"{submissions_str}"
    )


def download_submission(submission, group_arg, client):
    """Download the submission files.

    If there's one file, download it.  If there are multiple, then download to
    a directory.

    """
    # If the user provides a group argument like a uniqname or group pk, prefix
    # the filename with it.
    if group_arg:
        prefix = f"{group_arg}-submission-{submission['pk']}"
    else:
        prefix = f"submission-{submission['pk']}"

    # Download file to PWD.  If there are multiple files, put them in a new
    # directory.
    filenames = submission['submitted_filenames']
    if not filenames:
        sys.exit(
            "Error: no files to download for submission"
            f"{submission_str(submission)}"
        )
    elif len(filenames) == 1:
        filename = filenames[0]
        target = pathlib.Path(f"{prefix}-{filename}")
        download_file(filename, submission, target, client)
    else:
        dirname = pathlib.Path(prefix)
        if dirname.exists():
            sys.exit(f"Error: refuse to clobber directory: {dirname}")
        dirname.mkdir()
        for filename in filenames:
            download_file(filename, submission, dirname/filename, client)


def download_file(filename, submission, target, client):
    """Download the file named filename from submission pk submission.

    Save the file in path target/filename.
    """
    if target.exists():
        sys.exit(f"Error: refuse to clobber file: {target}")
    url = f"/api/submissions/{submission['pk']}/file/?filename={filename}"
    data = client.get(url)
    with target.open("wb") as targetfile:
        targetfile.write(data)
    print(f"Saved {target}")



"""Shared test fixtures."""
import json
import pytest


@pytest.fixture(name="constants")
def constants_setup():
    """Global constants."""
    return {
        "COURSE_109": {
            "pk": 109,
            "name": "EECS 485",
            "semester": "Spring",
            "year": 2021,
            "subtitle": "Web Systems",
            "num_late_days": 0,
            "allowed_guest_domain": "@umich.edu",
            "last_modified": "2021-04-07T02:19:22.818992Z"
        },
        "PROJECT_1005": {
            "pk": 1005,
            "name": "Project 1 - Templated Static Site Generator",
            "last_modified": "2021-05-13T19:46:38.254102Z",
            "course": 109,
            "visible_to_students": True,
            "closing_time": "2021-05-12T04:30:00Z",
            "soft_closing_time": "2021-05-12T03:59:00Z",
            "disallow_student_submissions": False,
            "disallow_group_registration": False,
            "guests_can_submit": True,
            "min_group_size": 1,
            "max_group_size": 1,
            "submission_limit_per_day": 3,
            "allow_submissions_past_limit": False,
            "groups_combine_daily_submissions": False,
            "submission_limit_reset_time": "00:00:00",
            "submission_limit_reset_timezone": "US/Eastern",
            "num_bonus_submissions": 0,
            "total_submission_limit": None,
            "allow_late_days": False,
            "ultimate_submission_policy": "best",
            "hide_ultimate_submission_fdbk": False,
            "instructor_files": [
                {
                    "pk": 21908,
                    "project": 1005,
                    "name": "autograder-6f82202d.tar.gz",
                    "last_modified": "2021-04-07T02:16:02.992523Z",
                    "size": 11323
                },
                {
                    "pk": 21909,
                    "project": 1005,
                    "name": "install",
                    "last_modified": "2021-04-07T02:16:03.147228Z",
                    "size": 2230
                }
            ],
            "expected_student_files": [
                {
                    "pk": 2545,
                    "project": 1005,
                    "pattern": "submit.tar.gz",
                    "min_num_matches": 1,
                    "max_num_matches": 1,
                    "last_modified": "2021-04-07T00:31:21.379258Z"
                }
            ],
            "has_handgrading_rubric": False,
            "send_email_on_submission_received": True,
            "send_email_on_non_deferred_tests_finished": True,
            "use_honor_pledge": True,
            "honor_pledge_text": (
                "I have neither given nor received aid on this project, "
                "nor have I concealed any violations of the Honor Code."
            )
        },
        "GROUP_246965": {
            "pk": 246965,
            "project": 1005,
            "extended_due_date": None,
            "member_names": [
                "awdeorio@umich.edu"
            ],
            "members": [
                {
                    "pk": 5,
                    "username": "awdeorio@umich.edu",
                    "first_name": "Andrew",
                    "last_name": "DeOrio",
                    "email": "",
                    "is_superuser": False
                }
            ],
            "bonus_submissions_remaining": 0,
            "late_days_used": {},
            "num_submissions": 2,
            "num_submits_towards_limit": 1,
            "created_at": "2021-04-21T17:01:37.807261Z",
            "last_modified": "2021-04-21T17:01:37.807025Z"
        },
        "SUBMISSION_1128572": {
            "pk": 1128572,
            "group": 246965,
            "timestamp": "2021-06-29T14:55:57.886137Z",
            "submitter": "awdeorio@umich.edu",
            "submitted_filenames": [
                "submit.tar.gz"
            ],
            "discarded_files": [],
            "missing_files": {},
            "status": "finished_grading",
            "is_past_daily_limit": False,
            "is_bonus_submission": False,
            "count_towards_total_limit": True,
            "does_not_count_for": [],
            "position_in_queue": 0,
            "grading_start_time": "2021-06-29T14:56:04.154353Z",
            "non_deferred_grading_end_time": "2021-06-29T14:56:56.289720Z",
            "last_modified": "2021-06-29T14:56:55.378590Z"
        },
        "SUBMISSION_1125717": {
            "pk": 1125717,
            "group": 246965,
            "timestamp": "2021-06-09T12:49:16.047791Z",
            "submitter": "awdeorio@umich.edu",
            "submitted_filenames": [
                "submit.tar.gz"
            ],
            "discarded_files": [],
            "missing_files": {},
            "status": "finished_grading",
            "is_past_daily_limit": False,
            "is_bonus_submission": False,
            "count_towards_total_limit": True,
            "does_not_count_for": [],
            "position_in_queue": 0,
            "grading_start_time": "2021-06-09T12:49:21.340598Z",
            "non_deferred_grading_end_time": "2021-06-09T12:50:26.978440Z",
            "last_modified": "2021-06-09T12:50:25.899462Z"
        }
    }


@pytest.fixture(name="api_mock")
def api_requests_mock(requests_mock, mocker, constants):
    """Mock Autograder API with hardcoded responses."""
    # Don't look for an API token on the filesystem
    mocker.patch("agiocli.api_client.get_api_token")

    # User
    requests_mock.get(
        "https://autograder.io/api/users/current/",
        headers={"Content-Type": "application/json"},
        text=json.dumps({
            "pk": 5,
            "username": "awdeorio@umich.edu",
            "first_name": "Andrew",
            "last_name": "DeOrio",
            "email": "",
            "is_superuser": False
        })
    )

    # Course list / admin
    requests_mock.get(
        "https://autograder.io/api/users/5/courses_is_admin_for/",
        headers={"Content-Type": "application/json"},
        text=json.dumps([
            {
                "pk": 111,
                "name": "EECS 280",
                "semester": "Spring",
                "year": 2021,
                "subtitle": "",
                "num_late_days": 0,
                "allowed_guest_domain": "@umich.edu",
                "last_modified": "2021-05-03T01:12:28.049482Z"
            },
            constants["COURSE_109"],
            {
                "pk": 129,
                "name": "EECS 485",
                "semester": "Fall",
                "year": 2021,
                "subtitle": "Web Systems",
                "num_late_days": 0,
                "allowed_guest_domain": "@umich.edu",
                "last_modified": "2021-06-23T13:54:07.942973Z"
            }
        ])
    )

    # Course list / staff
    requests_mock.get(
        "https://autograder.io/api/users/5/courses_is_staff_for/",
        headers={"Content-Type": "application/json"},
        text=json.dumps([
            {
                "pk": 1,
                "name": "EECS 280",
                "semester": "Fall",
                "year": 2016,
                "subtitle": "",
                "num_late_days": 0,
                "allowed_guest_domain": "@umich.edu",
                "last_modified": "2019-02-07T21:06:01.779838Z"
            },
        ])
    )

    # Course detail
    requests_mock.get(
        "https://autograder.io/api/courses/109/",
        headers={"Content-Type": "application/json"},
        text=json.dumps(constants["COURSE_109"]),
    )

    # Project list
    requests_mock.get(
        "https://autograder.io/api/courses/109/projects/",
        headers={"Content-Type": "application/json"},
        text=json.dumps([
            {
                "pk": 1007,
                "name": "Project 5 - Search Engine",
                "last_modified": "2021-06-21T16:42:49.476960Z",
                "course": 109,
                "visible_to_students": True,
                "closing_time": "2021-06-21T04:30:00Z",
                "soft_closing_time": "2021-06-21T03:59:00Z",
                "disallow_student_submissions": False,
                "disallow_group_registration": False,
                "guests_can_submit": False,
                "min_group_size": 1,
                "max_group_size": 3,
                "submission_limit_per_day": 3,
                "allow_submissions_past_limit": False,
                "groups_combine_daily_submissions": False,
                "submission_limit_reset_time": "00:00:00",
                "submission_limit_reset_timezone": "US/Eastern",
                "num_bonus_submissions": 0,
                "total_submission_limit": None,
                "allow_late_days": False,
                "ultimate_submission_policy": "best",
                "hide_ultimate_submission_fdbk": False,
                "instructor_files": [
                    {
                        "pk": 24749,
                        "project": 1007,
                        "name": "autograder-4d215e29.tar.xz",
                        "last_modified": "2021-06-14T03:39:17.838619Z",
                        "size": 4306452
                    },
                    {
                        "pk": 24750,
                        "project": 1007,
                        "name": "install",
                        "last_modified": "2021-06-14T03:39:18.125266Z",
                        "size": 2906
                    }
                ],
                "expected_student_files": [
                    {
                        "pk": 2547,
                        "project": 1007,
                        "pattern": "submit.tar.xz",
                        "min_num_matches": 1,
                        "max_num_matches": 1,
                        "last_modified": "2021-04-07T00:31:28.029083Z"
                    }
                ],
                "has_handgrading_rubric": False,
                "send_email_on_submission_received": True,
                "send_email_on_non_deferred_tests_finished": True,
                "use_honor_pledge": True,
                "honor_pledge_text": (
                    "I have neither given nor received aid on this project, "
                    "nor have I concealed any violations of the Honor Code."
                )
            },
            {
                "pk": 1008,
                "name": "Project 3 - Client-side Dynamic Pages",
                "last_modified": "2021-05-27T22:01:15.359868Z",
                "course": 109,
                "visible_to_students": True,
                "closing_time": "2021-05-26T04:30:00Z",
                "soft_closing_time": "2021-05-26T03:59:00Z",
                "disallow_student_submissions": False,
                "disallow_group_registration": False,
                "guests_can_submit": False,
                "min_group_size": 1,
                "max_group_size": 3,
                "submission_limit_per_day": 3,
                "allow_submissions_past_limit": False,
                "groups_combine_daily_submissions": False,
                "submission_limit_reset_time": "00:00:00",
                "submission_limit_reset_timezone": "US/Eastern",
                "num_bonus_submissions": 0,
                "total_submission_limit": None,
                "allow_late_days": False,
                "ultimate_submission_policy": "best",
                "hide_ultimate_submission_fdbk": False,
                "instructor_files": [
                    {
                        "pk": 22508,
                        "project": 1008,
                        "name": "autograder-bcc78f56-dirty.tar.gz",
                        "last_modified": "2021-05-18T22:00:36.641145Z",
                        "size": 168836
                    },
                    {
                        "pk": 21949,
                        "project": 1008,
                        "name": "install",
                        "last_modified": "2021-05-18T22:00:36.805560Z",
                        "size": 3057
                    }
                ],
                "expected_student_files": [
                    {
                        "pk": 2548,
                        "project": 1008,
                        "pattern": "submit.tar.gz",
                        "min_num_matches": 1,
                        "max_num_matches": 1,
                        "last_modified": "2021-04-07T00:31:30.640369Z"
                    }
                ],
                "has_handgrading_rubric": False,
                "send_email_on_submission_received": True,
                "send_email_on_non_deferred_tests_finished": True,
                "use_honor_pledge": True,
                "honor_pledge_text": (
                    "I have neither given nor received aid on this project, "
                    "nor have I concealed any violations of the Honor Code."
                )
            },
            constants["PROJECT_1005"],
            {
                "pk": 1009,
                "name": "Project 2 - Server-side Dynamic Pages",
                "last_modified": "2021-05-19T13:03:53.221672Z",
                "course": 109,
                "visible_to_students": True,
                "closing_time": "2021-05-19T04:30:00Z",
                "soft_closing_time": "2021-05-19T03:59:00Z",
                "disallow_student_submissions": False,
                "disallow_group_registration": False,
                "guests_can_submit": False,
                "min_group_size": 1,
                "max_group_size": 3,
                "submission_limit_per_day": 3,
                "allow_submissions_past_limit": False,
                "groups_combine_daily_submissions": False,
                "submission_limit_reset_time": "00:00:00",
                "submission_limit_reset_timezone": "US/Eastern",
                "num_bonus_submissions": 0,
                "total_submission_limit": None,
                "allow_late_days": False,
                "ultimate_submission_policy": "best",
                "hide_ultimate_submission_fdbk": False,
                "instructor_files": [
                    {
                        "pk": 22506,
                        "project": 1009,
                        "name": "autograder-788e5076.tar.gz",
                        "last_modified": "2021-05-11T19:21:55.373807Z",
                        "size": 282326
                    },
                    {
                        "pk": 22507,
                        "project": 1009,
                        "name": "install",
                        "last_modified": "2021-05-11T19:21:55.596257Z",
                        "size": 2095
                    }
                ],
                "expected_student_files": [
                    {
                        "pk": 2549,
                        "project": 1009,
                        "pattern": "submit.tar.gz",
                        "min_num_matches": 1,
                        "max_num_matches": 1,
                        "last_modified": "2021-04-07T00:31:33.584510Z"
                    }
                ],
                "has_handgrading_rubric": False,
                "send_email_on_submission_received": True,
                "send_email_on_non_deferred_tests_finished": True,
                "use_honor_pledge": True,
                "honor_pledge_text": (
                    "I have neither given nor received aid on this project, "
                    "nor have I concealed any violations of the Honor Code."
                )
            },
            {
                "pk": 1006,
                "name": "Project 4 - MapReduce",
                "last_modified": "2021-06-15T03:59:11.942724Z",
                "course": 109,
                "visible_to_students": True,
                "closing_time": "2021-06-14T04:30:00Z",
                "soft_closing_time": "2021-06-14T03:59:00Z",
                "disallow_student_submissions": False,
                "disallow_group_registration": False,
                "guests_can_submit": False,
                "min_group_size": 1,
                "max_group_size": 3,
                "submission_limit_per_day": 3,
                "allow_submissions_past_limit": False,
                "groups_combine_daily_submissions": False,
                "submission_limit_reset_time": "00:00:00",
                "submission_limit_reset_timezone": "US/Eastern",
                "num_bonus_submissions": 0,
                "total_submission_limit": None,
                "allow_late_days": False,
                "ultimate_submission_policy": "best",
                "hide_ultimate_submission_fdbk": False,
                "instructor_files": [
                    {
                        "pk": 21900,
                        "project": 1006,
                        "name": "autograder-8b2ec48d.tar.gz",
                        "last_modified": "2021-04-07T00:31:25.318107Z",
                        "size": 3077619
                    },
                    {
                        "pk": 21901,
                        "project": 1006,
                        "name": "install",
                        "last_modified": "2021-04-07T00:31:25.441090Z",
                        "size": 2098
                    }
                ],
                "expected_student_files": [
                    {
                        "pk": 2546,
                        "project": 1006,
                        "pattern": "submit.tar.gz",
                        "min_num_matches": 1,
                        "max_num_matches": 1,
                        "last_modified": "2021-04-07T00:31:25.503082Z"
                    }
                ],
                "has_handgrading_rubric": False,
                "send_email_on_submission_received": True,
                "send_email_on_non_deferred_tests_finished": True,
                "use_honor_pledge": True,
                "honor_pledge_text": (
                    "I have neither given nor received aid on this project, "
                    "nor have I concealed any violations of the Honor Code."
                )
            }
        ])
    )

    # Project detail
    requests_mock.get(
        "https://autograder.io/api/projects/1005/",
        headers={"Content-Type": "application/json"},
        text=json.dumps(constants["PROJECT_1005"]),
    )

    # Group list
    requests_mock.get(
        "https://autograder.io/api/projects/1005/groups/",
        headers={"Content-Type": "application/json"},
        text=json.dumps(
            [
                {
                    "pk": 243636,
                    "project": 1005,
                    "extended_due_date": None,
                    "member_names": [
                        "achitta@umich.edu"
                    ],
                    "members": [
                        {
                            "pk": 11854,
                            "username": "achitta@umich.edu",
                            "first_name": "Aditya",
                            "last_name": "Chitta",
                            "email": "",
                            "is_superuser": False
                        }
                    ],
                    "bonus_submissions_remaining": 0,
                    "late_days_used": {},
                    "num_submissions": 1,
                    "num_submits_towards_limit": 0,
                    "created_at": "2021-04-07T02:17:16.975983Z",
                    "last_modified": "2021-04-07T02:17:16.975700Z"
                },
                {
                    "pk": 246965,
                    "project": 1005,
                    "extended_due_date": None,
                    "member_names": [
                        "awdeorio@umich.edu"
                    ],
                    "members": [
                        {
                            "pk": 5,
                            "username": "awdeorio@umich.edu",
                            "first_name": "Andrew",
                            "last_name": "DeOrio",
                            "email": "",
                            "is_superuser": False
                        }
                    ],
                    "bonus_submissions_remaining": 0,
                    "late_days_used": {},
                    "num_submissions": 2,
                    "num_submits_towards_limit": 1,
                    "created_at": "2021-04-21T17:01:37.807261Z",
                    "last_modified": "2021-04-21T17:01:37.807025Z"
                },
            ]),
    )

    # Group detail
    requests_mock.get(
        "https://autograder.io/api/groups/246965/",
        headers={"Content-Type": "application/json"},
        text=json.dumps(constants["GROUP_246965"]),
    )

    # Submission list
    requests_mock.get(
        "https://autograder.io/api/groups/246965/submissions/",
        headers={"Content-Type": "application/json"},
        text=json.dumps([
            constants["SUBMISSION_1128572"],
            constants["SUBMISSION_1125717"],
        ])
    )

    # Submission detail
    requests_mock.get(
        "https://autograder.io/api/submissions/1128572/",
        headers={"Content-Type": "application/json"},
        text=json.dumps(constants["SUBMISSION_1128572"]),
    )

    # Ultimate submission detail
    requests_mock.get(
        "https://autograder.io/api/groups/246965/ultimate_submission/",
        headers={"Content-Type": "application/json"},
        text=json.dumps(constants["SUBMISSION_1125717"]),
    )



"""
System tests for the command line interface.

These tests use the Click testing interface.
https://click.palletsprojects.com/en/8.0.x/testing/
"""
import click
import click.testing
from agiocli.__main__ import main


def test_example():
    """Dummy example test."""
    runner = click.testing.CliRunner(mix_stderr=False)
    result = runner.invoke(main, ["--version"], catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "version" in result.output



"""System tests for courses subcommand.

These tests use the Click testing interface.
https://click.palletsprojects.com/en/8.0.x/testing/
"""
import json
import textwrap
import click
import click.testing
import freezegun
from agiocli.__main__ import main


# Unused arguments due to fixtures are endemic to pytest
# pylint: disable=unused-argument


def test_courses_list(api_mock):
    """Verify agio courses list option.

    $ agio courses --list

    api_mock is a shared test fixture that mocks responses to REST API
    requests.  It is implemented in conftest.py.

    """
    runner = click.testing.CliRunner()
    result = runner.invoke(main, ["courses", "--list"], catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert result.output == textwrap.dedent("""\
        [129] EECS 485 Fall 2021
        [109] EECS 485 Spring 2021
        [111] EECS 280 Spring 2021
        [1] EECS 280 Fall 2016
    """)


def test_courses_empty(api_mock, mocker):
    """Verify courses subcommand no input.

    $ agio courses

    api_mock is a shared test fixture that mocks responses to REST API
    requests.  It is implemented in conftest.py.

    """
    # Mock user-selection menu, users selects course pk 109
    course_109 = {
        'pk': 109,
        'name': 'EECS 485',
        'semester': 'Spring',
        'year': 2021,
        'subtitle': 'Web Systems',
        'num_late_days': 0,
        'allowed_guest_domain': '@umich.edu',
        'last_modified': '2021-04-07T02:19:22.818992Z'
    }
    mocker.patch("pick.pick", return_value=(course_109, 1))

    # Run agio, mocking the date to be Jun 2021.  We need to mock the date
    # because the prompt filters out past courses.
    # https://github.com/spulec/freezegun
    runner = click.testing.CliRunner()
    with freezegun.freeze_time("2021-06-15"):
        result = runner.invoke(main, ["courses"], catch_exceptions=False)

    # Check output
    assert result.exit_code == 0, result.output
    output_obj = json.loads(result.output)
    assert output_obj["pk"] == 109


def test_courses_pk(api_mock):
    """Verify courses subcommand with primary key input.

    $ agio courses 109

    api_mock is a shared test fixture that mocks responses to REST API
    requests.  It is implemented in conftest.py.

    """
    runner = click.testing.CliRunner()
    result = runner.invoke(main, ["courses", "109"], catch_exceptions=False)
    assert result.exit_code == 0, result.output
    output_obj = json.loads(result.output)
    assert output_obj["pk"] == 109


def test_courses_name(api_mock):
    """Verify courses subcommand with course name input.

    $ agio courses "EECS 485 Spring 2021"

    api_mock is a shared test fixture that mocks responses to REST API
    requests.  It is implemented in conftest.py.

    """
    runner = click.testing.CliRunner()
    result = runner.invoke(
        main, ["courses", "EECS 485 Spring 2021"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    output_obj = json.loads(result.output)
    assert output_obj["pk"] == 109


def test_courses_shortcut(api_mock):
    """Verify courses subcommand with shortcut input.

    $ agio courses eecs485sp21

    api_mock is a shared test fixture that mocks responses to REST API
    requests.  It is implemented in conftest.py.

    """
    runner = click.testing.CliRunner()
    result = runner.invoke(
        main, ["courses", "eecs485sp21"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    output_obj = json.loads(result.output)
    assert output_obj["pk"] == 109



"""
A command line interface to autograder.io.

Andrew DeOrio <awdeorio@umich.edu>
"""
import sys
import click
from agiocli import APIClient, TokenFileNotFound, utils


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option()
@click.option("-d", "--debug", is_flag=True, help="Debug output")
@click.pass_context
def main(ctx, debug):
    """Autograder.io command line interface."""
    # Pass global flags to subcommands via Click context
    # https://click.palletsprojects.com/en/latest/commands/#nested-handling-and-contexts
    ctx.ensure_object(dict)
    ctx.obj["DEBUG"] = debug


@main.command()
@click.pass_context
def login(ctx):
    """Show current authenticated user."""
    try:
        client = APIClient.make_default(debug=ctx.obj["DEBUG"])
    except TokenFileNotFound as err:
        sys.exit(err)
    user = client.get("/api/users/current/")
    print(f"{user['username']} {user['first_name']} {user['last_name']}")


@main.command()
@click.argument("course_arg", required=False)
@click.option("-l", "--list", "show_list", is_flag=True,
              help="List courses and exit.")
@click.option("-w", "--web", is_flag=True, help="Open course in browser.")
@click.pass_context
# The \b character in the docstring prevents Click from rewraping a paragraph.
# We need to tell pycodestyle to ignore it.
# https://click.palletsprojects.com/en/8.0.x/documentation/#preventing-rewrapping
def courses(ctx, course_arg, show_list, web):  # noqa: D301
    """Show course detail or list courses.

    COURSE_ARG is a primary key, name, or shorthand.

    \b
    EXAMPLES:
    agio courses --list
    agio courses
    agio courses 109
    agio courses eecs485sp21

    """
    try:
        client = APIClient.make_default(debug=ctx.obj["DEBUG"])
    except TokenFileNotFound as err:
        sys.exit(err)

    # Handle --list: list courses and exit
    if show_list:
        course_list = utils.get_current_course_list(client)
        for i in course_list:
            print(utils.course_str(i))
        return

    # Select a course and print or open it
    course = utils.get_course_smart(course_arg, client)
    if web:
        utils.open_web(f"https://autograder.io/web/course/{course['pk']}")
        return
    print(utils.dict_str(course))


@main.command()
@click.argument("project_arg", required=False)
@click.option("-c", "--course", "course_arg",
              help="Course pk, name, or shorthand.")
@click.option("-l", "--list", "show_list", is_flag=True,
              help="List projects and exit.")
@click.option("-w", "--web", is_flag=True, help="Open project in browser.")
@click.pass_context
# The \b character in the docstring prevents Click from rewraping a paragraph.
# We need to tell pycodestyle to ignore it.
# https://click.palletsprojects.com/en/8.0.x/documentation/#preventing-rewrapping
def projects(ctx, project_arg, course_arg, show_list, web):  # noqa: D301
    """Show project detail or list projects.

    PROJECT_ARG is a primary key, name, or shorthand.

    \b
    EXAMPLES:
    agio projects --list
    agio projects
    agio projects 1005
    agio projects --course 109 p1
    agio projects --course eecs485sp21 p1
    agio projects p1

    """
    try:
        client = APIClient.make_default(debug=ctx.obj["DEBUG"])
    except TokenFileNotFound as err:
        sys.exit(err)

    # Handle --list: list projects and exit
    if show_list:
        course = utils.get_course_smart(course_arg, client)
        project_list = utils.get_course_project_list(course, client)
        for i in project_list:
            print(utils.project_str(i))
        return

    # Select a project and print or open it
    project = utils.get_project_smart(project_arg, course_arg, client)
    if web:
        utils.open_web(f"https://autograder.io/web/project/{project['pk']}")
        return
    print(utils.dict_str(project))


@main.command()
@click.argument("group_arg", required=False)
@click.option("-c", "--course", "course_arg",
              help="Course pk, name, or shorthand.")
@click.option("-p", "--project", "project_arg",
              help="Project pk, name, or shorthand.")
@click.option("-l", "--list", "show_list", is_flag=True,
              help="List groups and exit.")
@click.option("-w", "--web", is_flag=True, help="Open group in browser.")
@click.pass_context
# The \b character in the docstring prevents Click from rewraping a paragraph.
# We need to tell pycodestyle to ignore it.
# https://click.palletsprojects.com/en/8.0.x/documentation/#preventing-rewrapping
def groups(ctx, group_arg, project_arg, course_arg, show_list, web):  # noqa: D301
    """Show group detail or list groups.

    GROUP_ARG is a primary key, name, or member uniqname.

    \b
    EXAMPLES:
    agio groups --list
    agio groups
    agio groups 246965
    agio groups awdeorio
    agio groups awdeorio --project 1005
    agio groups awdeorio --course eecs485sp21 --project p1

    """
    # We must have an function argument for each CLI argument or option
    # pylint: disable=too-many-arguments

    try:
        client = APIClient.make_default(debug=ctx.obj["DEBUG"])
    except TokenFileNotFound as err:
        sys.exit(err)

    # Handle --list: list groups and exit
    if show_list:
        project = utils.get_project_smart(project_arg, course_arg, client)
        group_list = utils.get_group_list(project, client)
        for i in group_list:
            print(utils.group_str(i))
        return

    # Select a group and print or open it
    group = utils.get_group_smart(group_arg, project_arg, course_arg, client)
    if web:
        utils.open_web(
            "https://autograder.io/web/"
            f"project/{group['project']}"
            f"?current_tab=student_lookup"
            f"&current_student_lookup={group['pk']}"
        )
        return
    print(utils.dict_str(group))


@main.command()
@click.argument("submission_arg", required=False)
@click.option("-c", "--course", "course_arg",
              help="Course pk, name, or shorthand.")
@click.option("-p", "--project", "project_arg",
              help="Project pk, name, or shorthand.")
@click.option("-g", "--group", "group_arg",
              help="Group pk or member uniqname.")
@click.option("-l", "--list", "show_list", is_flag=True,
              help="List groups and exit.")
@click.option("-d", "--download", is_flag=True,
              help="Download submission files.")
@click.pass_context
# The \b character in the docstring prevents Click from rewraping a paragraph.
# We need to tell pycodestyle to ignore it.
# https://click.palletsprojects.com/en/8.0.x/documentation/#preventing-rewrapping
def submissions(ctx, submission_arg, group_arg,
                project_arg, course_arg, show_list, download):  # noqa: D301
    """Show submission detail or list submissions.

    SUBMISSION_ARG is a primary key, 'best', or 'last'

    \b
    EXAMPLES:
    agio submissions --list
    agio submissions
    agio submissions 1128572
    agio submissions --course eecs485sp21 --project p1 --group awdeorio
    agio submissions [...] best
    agio submissions [...] last
    agio submissions [...] --download
    """
    # We must have an function argument for each CLI argument or option
    # pylint: disable=too-many-arguments

    try:
        client = APIClient.make_default(debug=ctx.obj["DEBUG"])
    except TokenFileNotFound as err:
        sys.exit(err)

    # Handle --list: list submissions and exit
    if show_list:
        group = utils.get_group_smart(
            group_arg, project_arg, course_arg, client
        )
        submission_list = utils.get_submission_list(group, client)
        for i in submission_list:
            print(utils.submission_str(i))
        return

    # Select a submission
    submission = utils.get_submission_smart(
        submission_arg, group_arg, project_arg, course_arg, client
    )

    # Handle --download: download the submission and exit
    if download:
        utils.download_submission(submission, group_arg, client)
        return

    # Default: print submission
    print(utils.dict_str(submission))


if __name__ == "__main__":
    # These errors are endemic to click
    # pylint: disable=no-value-for-parameter,unexpected-keyword-arg
    main(obj={})



"""System tests for groups subcommand.

These tests use the Click testing interface.
https://click.palletsprojects.com/en/8.0.x/testing/
"""
import json
import click
import click.testing
import freezegun
from agiocli.__main__ import main


# Unused arguments due to fixtures are endemic to pytest
# pylint: disable=unused-argument


def test_groups_list(api_mock):
    """Verify agio groups list option when project is specified.

    $ agio groups --list --project 1005

    api_mock is a shared test fixture that mocks responses to REST API
    requests.  It is implemented in conftest.py.

    """
    runner = click.testing.CliRunner()
    result = runner.invoke(
        main, [
            "groups",
            "--list",
            "--project", "1005",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    assert "[243636] achitta" in result.output
    assert "[246965] awdeorio" in result.output


def test_groups_pk(api_mock):
    """Verify groups subcommand with primary key input.

    $ agio groups 246965

    api_mock is a shared test fixture that mocks responses to REST API
    requests.  It is implemented in conftest.py.

    """
    runner = click.testing.CliRunner()
    result = runner.invoke(main, ["groups", "246965"], catch_exceptions=False)
    assert result.exit_code == 0, result.output
    output_obj = json.loads(result.output)
    assert output_obj["pk"] == 246965
    assert output_obj["members"] == [
        {
            'email': '',
            'first_name': 'Andrew',
            'is_superuser': False,
            'last_name': 'DeOrio',
            'pk': 5,
            'username': 'awdeorio@umich.edu',
        },
    ]


def test_groups_uniqname(api_mock):
    """Verify groups subcommand with group member uniqname input.

    $ agio groups \
        --course eecs485sp21 \
        --project p1 \
        awdeorio

    api_mock is a shared test fixture that mocks responses to REST API
    requests.  It is implemented in conftest.py.

    """
    runner = click.testing.CliRunner()
    result = runner.invoke(
        main, [
            "groups",
            "--course", "eecs485sp21",
            "--project", "p1",
            "awdeorio",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.output
    output_obj = json.loads(result.output)
    assert output_obj["pk"] == 246965  # awdeorio's group


def test_groups_empty(api_mock, mocker, constants):
    """Verify groups subcommand no input.

    $ agio groups

    api_mock is a shared test fixture that mocks responses to REST API
    requests.  It is implemented in conftest.py.

    """
    # Mock user-selection menu, users selects course 109, then project 1005.
    # These are constants in conftest.py.  Mock input "awdeorio", which selects
    # a group.
    mocker.patch("pick.pick", side_effect=[
        (constants["COURSE_109"], 1),  # First call to pick() selects course
        (constants["PROJECT_1005"], 0),  # Second call selects project
    ])
    mocker.patch("builtins.input", return_value="awdeorio")

    # Run agio, mocking the date to be Jun 2021.  We need to mock the date
    # because the prompt filters out past courses.
    # https://github.com/spulec/freezegun
    runner = click.testing.CliRunner()
    with freezegun.freeze_time("2021-06-15"):
        result = runner.invoke(main, ["groups"], catch_exceptions=False)

    # Check output
    assert result.exit_code == 0, result.output
    output_obj = json.loads(result.output)
    assert output_obj["pk"] == 246965  # awdeorio's group



"""Autograder.io CLI build and install configuration."""
import os
import io
import setuptools


# Read the contents of README file
PROJECT_DIR = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(PROJECT_DIR, "README.md"), encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()


setuptools.setup(
    name="agiocli",
    description="A command line interface to autograder.io",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    version="0.2.0",
    author="Andrew DeOrio",
    author_email="awdeorio@umich.edu",
    url="https://github.com/eecs485staff/agio-cli/",
    license="MIT",
    packages=["agiocli"],
    keywords=[
        "autograder", "autograder.io", "auto grader",
        "agcli", "agio", "ag-cli", "agio-cli",
    ],
    install_requires=[
        "click",
        "pick",
        "python-dateutil",
        "requests",
    ],
    extras_require={
        "dev": [
            "pdbpp",
            "twine",
            "tox",
        ],
        "test": [
            "check-manifest",
            "freezegun",
            "pycodestyle",
            "pydocstyle",
            "pylint",
            "pytest",
            "pytest-cov",
            "pytest-mock",
            "requests-mock",
        ],
    },
    python_requires='>=3.6',
    entry_points={
        "console_scripts": [
            "agio = agiocli.__main__:main",
        ]
    },
)



"""Unit tests for smart user input string matching."""
import pytest
from agiocli import utils


@pytest.mark.parametrize(
    "search, expected_course_pk",
    [
        ("EECS 280 Spring 2021", 111),
        ("eecs 280 spring 2021", 111),
        ("EECS 485 Spring 2021", 109),
        ("EECS 280 Spring 21", 111),
        ("EECS 485 Spring 21", 109),
        ("EECS 280 sp 21", 111),
        ("EECS 485 sp 21", 109),
        ("280 sp 21", 111),
        ("485 sp 21", 109),
        ("eecs280sp21", 111),
        ("eecs485sp21", 109),
        ("eecs485-sp-21", 109),
        ("eecs485-sp_21", 109),
    ]
)
def test_course_match_input_patterns(search, expected_course_pk):
    """Many supported input patterns."""
    matches = utils.course_match(search, COURSES)
    assert len(matches) == 1
    course = matches[0]
    assert course["pk"] == expected_course_pk


@pytest.mark.parametrize(
    "search, expected_course_pk",
    [
        ("EECS 598 Winter 2019", 38),
        ("EECS 398 Fall 2019", 53),
        ("EECS 490 Winter 2020", 67),
        # ("EECS 484 Fall 2020", 77)
        # Above won't match because there is a duplicate
        ("EECS 498 Fall 2020", 79),  # This won't match on EECS 598
        ("EECS 398 Winter 2021", 90),
        ("EECS 598 wn 19", 38),
        ("EECS 398 fa 19", 53),
        ("EECS 490 wn 20", 67),
        # ("EECS 484 f 20", 77),
        # Above won't match because there is a duplicate
        ("EECS 498 f 20", 79),  # This won't match on EECS 598
        ("EECS 398 w 21", 90),
        ("598 wn 19", 38),
        ("398 fa 19", 53),
        ("490 wn 20", 67),
        # ("484 f 20", 77),
        # Above won't match because there is a duplicate
        ("498 f 20", 79),  # This won't match on EECS 598
        ("398 w 21", 90),
        ("598wn19", 38),
        ("398fa19", 53),
        ("490wn20", 67),
        # ("484f20", 77),
        # Above won't match because there is a duplicate
        ("498f20", 79),  # This won't match on EECS 598
        ("398w21", 90),
    ]
)
def test_course_match_pattern(search, expected_course_pk):
    """Many supported input patterns."""
    matches = utils.course_match(search, COURSES)
    assert len(matches) == 1
    course = matches[0]
    assert course["pk"] == expected_course_pk


@pytest.mark.parametrize(
    "search",
    [
        ("EECS 280 Spring 2016"),
        ("EECS 485 Spring 2016"),
        ("EECS 280 Spring 16"),
        ("EECS 485 Spring 16"),
        ("EECS 280 sp 16"),
        ("EECS 485 sp 16"),
        ("280 sp 16"),
        ("485 sp 16"),
        ("eecs280sp16"),
        ("eecs485sp16"),
    ]
)
def test_course_match_bad_year(search):
    """Bad year in pattern."""
    matches = utils.course_match(search, COURSES)
    assert not matches


@pytest.mark.parametrize(
    "search, expected_project_pk",
    [
        ("Lab 01 - Getting Started", 435),
        ("Lab 1 - Getting Started", 435),
        ("Lab 1", 435),
        ("lab 1", 435),
        ("lab 01", 435),
        ("Lab01", 435),
        ("lab1", 435),
        ("l1", 435),
        ("L1", 435),
        ("Project 2 - Images", 423),
        ("project 2 - images", 423),
        ("Images", 423),
        ("Project 2", 423),
        ("Project2", 423),
        ("Project 02", 423),
        ("Project02", 423),
        ("p2", 423),
        ("P2", 423),
        ("p02", 423),
        ("P02", 423),
        ("Project-2", 423),
        ("Project-02", 423),
        ("p-2", 423),
        ("P-2", 423),
        ("p-02", 423),
        ("P-02", 423),
        ("P_02", 423),
        ("P_02_images", 423),
        ("Polymorphism", 424),
    ]
)
def test_project_match_pattern(search, expected_project_pk):
    """Many supported input patterns."""
    matches = utils.project_match(search, PROJECTS)
    assert len(matches) == 1
    project = matches[0]
    assert project["pk"] == expected_project_pk


@pytest.mark.parametrize(
    "search",
    [
        ("Lab27"),
        ("P9"),
        ("Project 9"),
        ("Project 09"),
        ("Project 1 Polymorphism"),
        ("Project 3 - Images"),
        ("Lab 2 - Images"),
        ("L14"),
    ]
)
def test_project_match_bad_num(search):
    """Bad project number in pattern."""
    matches = utils.project_match(search, PROJECTS)
    assert not matches


COURSES = [
    {"pk": 1, "name": "EECS 280", "semester": "Fall", "year": 2016},
    {"pk": 2, "name": "EECS 280", "semester": "Winter", "year": 2017},
    {"pk": 3, "name": "EECS 490", "semester": "Fall", "year": 2017},
    {"pk": 4, "name": "EECS 280", "semester": "Spring", "year": 2017},
    {"pk": 5, "name": "ENGR 101 F17", "semester": None, "year": None},
    {"pk": 7, "name": "EECS 280", "semester": "Fall", "year": 2017},
    {"pk": 8, "name": "EECS 485", "semester": "Fall", "year": 2017},
    {"pk": 9, "name": "AERO 552 Practice", "semester": None, "year": None},
    {"pk": 10, "name": "EECS 481", "semester": "Winter", "year": 2018},
    {"pk": 12, "name": "EECS 280", "semester": "Winter", "year": 2018},
    {"pk": 14, "name": "ENGR 101", "semester": "Winter", "year": 2018},
    {"pk": 15, "name": "EECS 485", "semester": "Winter", "year": 2018},
    {"pk": 16, "name": "EECS 398: System Design of a Search Engine",
        "semester": "Winter", "year": 2019},
    {"pk": 17, "name": "EECS 280", "semester": "Spring", "year": 2018},
    {"pk": 18, "name": "EECS 370", "semester": "Spring", "year": 2018},
    {"pk": 19, "name": "Demo EECS 280 F17", "semester": None, "year": None},
    {"pk": 20, "name": "Demo Course", "semester": None, "year": None},
    {"pk": 21, "name": "EECS 280 Diagnostic", "semester": None, "year": None},
    {"pk": 22, "name": "EECS 183", "semester": "Fall", "year": 2018},
    {"pk": 23, "name": "Copy of EECS 280 SP 2018",
        "semester": None, "year": None},
    {"pk": 25, "name": "EECS 490", "semester": "Fall", "year": 2018},
    {"pk": 26, "name": "EECS 280", "semester": "Fall", "year": 2018},
    {"pk": 27, "name": "EECS 285", "semester": "Fall", "year": 2018},
    {"pk": 29, "name": "EECS 485", "semester": "Fall", "year": 2018},
    {"pk": 30, "name": "ENGR 101", "semester": "Fall", "year": 2018},
    {"pk": 31, "name": "EECS 498-001 (Data Mining)",
        "semester": "Fall", "year": 2018},
    {"pk": 32, "name": "EECS 280", "semester": "Winter", "year": 2019},
    {"pk": 33, "name": "EECS 183 All Projects",
        "semester": None, "year": None},
    {"pk": 34, "name": "EECS 183", "semester": "Winter", "year": 2019},
    {"pk": 35, "name": "EECS 485", "semester": "Winter", "year": 2019},
    {"pk": 36, "name": "ENGR 101", "semester": "Winter", "year": 2019},
    {"pk": 37, "name": "EECS 481", "semester": "Winter", "year": 2019},
    {"pk": 38, "name": "EECS 598-008 Advanced Data Mining",
        "semester": "Winter", "year": 2019},
    {"pk": 40, "name": "EECS 483", "semester": "Winter", "year": 2019},
    {"pk": 41, "name": "EECS 493", "semester": "Winter", "year": 2019},
    {"pk": 42, "name": "EECS 370", "semester": "Fall", "year": 2019},
    {"pk": 43, "name": "EECS 280", "semester": "Spring", "year": 2019},
    {"pk": 44, "name": "EECS 490", "semester": "Fall", "year": 2019},
    {"pk": 45, "name": "EECS 285", "semester": "Fall", "year": 2019},
    {"pk": 46, "name": "EECS 485", "semester": "Fall", "year": 2019},
    {"pk": 47, "name": "EECS 183", "semester": "Fall", "year": 2019},
    {"pk": 48, "name": "ENGR 151", "semester": "Fall", "year": 2019},
    {"pk": 49, "name": "MSci Computer Science",
        "semester": "Summer", "year": 2020},
    {"pk": 50, "name": "EECS 280", "semester": "Fall", "year": 2019},
    {"pk": 51, "name": "EECS 481", "semester": "Fall", "year": 2019},
    {"pk": 52, "name": "ENGR 101", "semester": "Fall", "year": 2019},
    {"pk": 53, "name": "EECS 398: System Design of a Search Engine",
        "semester": "Fall", "year": 2019},
    {"pk": 54, "name": "EECS 484", "semester": "Fall", "year": 2019},
    {"pk": 55, "name": "EECS 183", "semester": "Winter", "year": 2020},
    {"pk": 56, "name": "EECS 280", "semester": "Winter", "year": 2020},
    {"pk": 57, "name": "ENGR 101", "semester": "Winter", "year": 2020},
    {"pk": 58, "name": "EECS 481", "semester": "Winter", "year": 2020},
    {"pk": 59, "name": "EECS 485", "semester": "Winter", "year": 2020},
    {"pk": 60, "name": "EECS 370", "semester": "Winter", "year": 2020},
    {"pk": 61, "name": "CIS 350", "semester": "Winter", "year": 2020},
    {"pk": 62, "name": "EECS 476", "semester": "Winter", "year": 2020},
    {"pk": 63, "name": "EECS 484", "semester": "Winter", "year": 2020},
    {"pk": 64, "name": "EECS 483", "semester": "Winter", "year": 2020},
    {"pk": 67, "name": "EECS 490 (Omar)", "semester": "Winter", "year": 2020},
    {"pk": 68, "name": "EECS 481", "semester": "Spring", "year": 2020},
    {"pk": 69, "name": "EECS 370", "semester": "Spring", "year": 2020},
    {"pk": 70, "name": "EECS 280", "semester": "Spring", "year": 2020},
    {"pk": 71, "name": "EECS 484", "semester": "Spring", "year": 2020},
    {"pk": 74, "name": "EECS 485", "semester": "Summer", "year": 2020},
    {"pk": 75, "name": "EECS 280 - INCOMPLETES",
        "semester": "Winter", "year": 2020},
    {"pk": 76, "name": "EECS 484 (copied from F19)",
        "semester": "Fall", "year": 2020},
    {"pk": 77, "name": "EECS 484 Fall 2020 (official)",
        "semester": "Fall", "year": 2020},
    {"pk": 78, "name": "EECS 183", "semester": "Fall", "year": 2020},
    {"pk": 79, "name": "EECS 498/598 Deep Learning for Computer Vision",
        "semester": "Fall", "year": 2020},
    {"pk": 80, "name": "EECS 280", "semester": "Fall", "year": 2020},
    {"pk": 81, "name": "CIS 350", "semester": "Fall", "year": 2020},
    {"pk": 82, "name": "EECS 481", "semester": "Fall", "year": 2020},
    {"pk": 83, "name": "EECS 370", "semester": "Fall", "year": 2020},
    {"pk": 85, "name": "EECS 485", "semester": "Fall", "year": 2020},
    {"pk": 86, "name": "EECS 285", "semester": "Fall", "year": 2020},
    {"pk": 87, "name": "ENGR 101", "semester": "Fall", "year": 2020},
    {"pk": 88, "name": "Research and Development Autograder",
        "semester": "Fall", "year": 2020},
    {"pk": 89, "name": "ENGR 151", "semester": "Fall", "year": 2020},
    {"pk": 90, "name": "EECS 398-001", "semester": "Winter", "year": 2021},
    {"pk": 91, "name": "EECS 183 - Elevators Development OUTDATED",
        "semester": "Fall", "year": 2020},
    {"pk": 93, "name": "ENGR 101", "semester": "Winter", "year": 2021},
    {"pk": 94, "name": "EECS 485", "semester": "Winter", "year": 2021},
    {"pk": 95, "name": "EECS 440: System Design of a Search Engine",
        "semester": "Winter", "year": 2021},
    {"pk": 96, "name": "EECS 481", "semester": "Winter", "year": 2021},
    {"pk": 97, "name": "EECS 484 Winter 2021",
        "semester": "Winter", "year": 2021},
    {"pk": 98, "name": "EECS 476", "semester": "Winter", "year": 2021},
    {"pk": 99, "name": "EECS 183", "semester": "Winter", "year": 2021},
    {"pk": 100, "name": "EECS 280", "semester": "Winter", "year": 2021},
    {"pk": 101, "name": "EECS 370", "semester": "Winter", "year": 2021},
    {"pk": 102, "name": "EECS 590", "semester": "Winter", "year": 2021},
    {"pk": 106, "name": "EECS 442 Computer Vision Winter 2021",
        "semester": "Winter", "year": 2021},
    {"pk": 107, "name": "EECS 483", "semester": "Winter", "year": 2021},
    {"pk": 108, "name": "CIS 350", "semester": "Winter", "year": 2021},
    {"pk": 109, "name": "EECS 485", "semester": "Spring", "year": 2021},
    {"pk": 110, "name": "EECS 484 Spring 2021",
        "semester": "Spring", "year": 2021},
    {"pk": 111, "name": "EECS 280", "semester": "Spring", "year": 2021},
    {"pk": 112, "name": "EECS 370", "semester": "Spring", "year": 2021},
    {"pk": 113, "name": "EECS 370 - Make up",
        "semester": "Spring", "year": 2021},
    {"pk": 126, "name": "ENGR 101", "semester": "Fall", "year": 2021}
]


# These projects are from EECS 280 Fall 2019
# GET https://autograder.io/api/courses/50/projects/
# I removed the instructor_files and expected_student_files for brevity
PROJECTS = [
    {"pk": 424, "name": "Lab 05 - Inheritance and Subtype Polymorphism"},
    {"pk": 422, "name": "Lab 08 - The Big Three"},
    {"pk": 425, "name": "Lab 03 - Strings and I/O"},
    {"pk": 423, "name": "Project 2 - Images"},
    {"pk": 427, "name": "Project 3 - Euchre"},
    {"pk": 435, "name": "Lab 01 - Getting Started"},
    {"pk": 430, "name": "Project 1 - Statistics"},
    {"pk": 429, "name": "Lab 09 - Recursion"},
    {"pk": 432, "name": "Lab 10 - Functors"},
    {"pk": 428, "name": "Lab 07 - Dynamic Memory"},
    {"pk": 436, "name": "Project 4 - Web"},
    {"pk": 433, "name": "Lab 04 - Abstract Data Types"},
    {"pk": 431, "name": "Lab 02 - Arrays and Pointers"},
    {"pk": 434, "name": "Project 5 - Machine Learning"},
    {"pk": 426, "name": "Lab 06 - Container ADTs"},
]
